# Copyright (c) 2007, Libero Spagnolini
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the authors nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# Modified for Tcl3D by Paul Obermeier 2007/04/14
# See www.tcl3d.org for the Tcl3D extension.
#
# The demo has been modified to allow up to 2 parameters to be changed
# interactively via a slider.
# The parameter range of the two sliders can be provided as comment lines
# at the top of the shader source files.
# Further enhancements include:
# Loading of image files of any size via the "Load image" button. All image files
# with an extension of .jpg in the directory of the script are automatically recognized
# and inserted into the "Images" labelframe.
# Add your own shader without modifying the Tcl script by adding a new file with extension 
# .frag in the directory of the script.
#
# A description of the effect shaders and the original sources are
# available at http://dem.ocracy.org/libero/photobooth/

package require tcl3d 0.3.2
package require Img

# texture stuff
set sizeTexX 512
set sizeTexY 512
set g_tex [tcl3dVector GLuint 1]

set g_lastDir [pwd]

# Take care, of the script is running from within a Starpack.
set pathName [file dirname [info script]]
set realPath [tcl3dGetExtFile $pathName]

# Get a list of all available shader programs.
set g_Effects [lsort [glob -directory $realPath "*.frag"]]

# Get a list of all available texture images. Must be 512x512.
set g_Textures [lsort [glob -directory $realPath "*.jpg"]]

# Variables for holding current settings.
set g_current(Texture) [lindex $g_Textures 0]
set g_current(Shader)  [lindex $g_Effects 0]
set g_current(Param,1) 0
set g_current(Param,2) 0

# Show errors occuring in the Togl callbacks.
proc bgerror { msg } {
    tk_messageBox -icon error -type ok -message "Error: $msg"
    exit
}

# Print info message into widget a the bottom of the window.
proc PrintInfo { msg } {
    if { [winfo exists .fr.info] } {
        .fr.info configure -text $msg
    }
}

# debug glsl
proc PrintInfoLog { obj } {
    set infologLength [tcl3dVector GLint 1]
    glGetObjectParameterivARB $obj GL_OBJECT_INFO_LOG_LENGTH_ARB $infologLength
    set infoLen [$infologLength get 0]
    if { $infoLen > 1 } {
        set infoLog [tcl3dVector GLubyte $infoLen]
        glGetInfoLogARB $obj $infologLength "NULL" $infoLog
        puts [tcl3dVectorToString $infoLog $infoLen]
    }
    $infologLength delete
}

proc LoadFileIntoTextWidget { w fileName } {
    set retVal [catch {open $fileName r} fp]
    if { $retVal != 0 } {
        error "Could not open file $fileName for reading."
    }
    $w delete 1.0 end
    while { ![eof $fp] } {
        $w insert end [read $fp 2048]
    }
    close $fp
}

proc TextWidgetToFile { w fileName } {
    set fileTypes {
            {"Tcl files" ".tcl"}
            {"All files" "*"} }

    if { [info exists      ::starkit::topdir] && \
         [file isdirectory $::starkit::topdir] } {
        set dumpDir [file dirname $::starkit::topdir]
    } else {
        set dumpDir [file dirname $fileName]
    }
    set dumpFile [file tail $fileName]
    set saveName [tk_getSaveFile -filetypes $fileTypes \
                          -initialfile $dumpFile -initialdir $dumpDir \
                          -title "Save Tcl script to file system"]
    if { $saveName != "" } {
        set retVal [catch {open $saveName w} fp]
        if { $retVal != 0 } {
            error "Could not open file $saveName for writing."
        }
        puts -nonewline $fp [$w get 1.0 end]
        close $fp
    }
}

proc AddMenuCmd { menu label acc cmd } {
    $menu add command -label $label -accelerator $acc -command $cmd
}

proc ShowEditor { fileName } {
    global gPres

    # Font to be used in the text widget.
    set textFont {-family {Courier} -size 10}

    if { ! [info exists gPres(editCount)] } {
        set gPres(editCount) 0
    } else {
        incr gPres(editCount)
    }

    set titleStr "GLSL shader: [file tail $fileName]"
    set tw ".poTextEdit_$gPres(editCount)"

    toplevel $tw
    wm title $tw $titleStr
    frame $tw.workfr -relief sunken -borderwidth 1
    pack  $tw.workfr -side top -fill both -expand 1

    set hMenu $tw.menufr
    menu $hMenu -borderwidth 2 -relief sunken
    $hMenu add cascade -menu $hMenu.file -label File -underline 0

    set textId [tcl3dCreateScrolledText $tw.workfr "$fileName" -font $textFont]

    set fileMenu $hMenu.file
    menu $fileMenu -tearoff 0
    AddMenuCmd $fileMenu "Save as ..." "Ctrl+S" "TextWidgetToFile $textId [list $fileName]"
    AddMenuCmd $fileMenu "Close" "Ctrl+W" "destroy $tw"

    bind $tw <Control-s> "TextWidgetToFile $textId [list $fileName]"
    bind $tw <Control-w> "destroy $tw"
    bind $tw <Escape>    "destroy $tw"
    wm protocol $tw WM_DELETE_WINDOW "destroy $tw"

    $tw configure -menu $hMenu 

    LoadFileIntoTextWidget $textId $fileName
    $textId configure -state disabled -cursor top_left_arrow
    focus $tw
}

proc ViewShaderFile {} {
    ShowEditor $::g_current(Shader)
}

proc AskOpen {} {
    global gPo

    set fileTypes {
        { "All files" * }
    }
    set imgName [tk_getOpenFile -filetypes $fileTypes \
                                -initialdir $::g_lastDir]
    if { $imgName != "" } {
        set ::g_lastDir [file dirname $imgName]
        ReadImg $imgName
        AddTexToList $imgName
        set ::g_current(Texture) $imgName
    }
}

proc GetBestSquare { w h } {
    if { $w > $h } {
        set val $w
    } else {
        set val $h
    }
    set sqrList { 1 2 4 8 16 32 64 128 256 512 1024 2048 4096 8192 16384 }
    foreach sqr $sqrList {
        if { $val <= $sqr } {
            return $sqr
        }
    }
}

proc ReadImg { filename } {
    set retVal [catch {set phImg [image create photo -file $filename]} err1]
    if { $retVal != 0 } {
        error "Error reading image $filename ($err1)"
    } else {
        set w [image width  $phImg]
        set h [image height $phImg]
        set n [tcl3dPhotoChans $phImg]

        set sqr [GetBestSquare $w $h]
        set ::g_texScaleS [expr double ($w) / $sqr]
        set ::g_texScaleT [expr double ($h) / $sqr]
        set sqrImg [image create photo -width $sqr -height $sqr]
        $sqrImg copy $phImg -from 0 0 $w $h -to 0 [expr $sqr -$h]
        set texData [tcl3dVector GLubyte [expr {$sqr * $sqr * $n}]]
        tcl3dPhoto2Vector $sqrImg $texData
        image delete $phImg
        image delete $sqrImg
    }

    # generate a "good looking" texture
    glGenTextures 1 $::g_tex
    glBindTexture GL_TEXTURE_2D [$::g_tex get 0]

    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_S $::GL_CLAMP
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_WRAP_T $::GL_CLAMP
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MAG_FILTER $::GL_LINEAR
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER $::GL_LINEAR

    if {$n == 4 } {
        glTexImage2D GL_TEXTURE_2D 0 $::GL_RGBA \
                     $sqr $sqr 0 GL_RGBA GL_UNSIGNED_BYTE $texData
    } else {
        glTexImage2D GL_TEXTURE_2D 0 $::GL_RGB \
                     $sqr $sqr 0 GL_RGB GL_UNSIGNED_BYTE $texData
    }

    # clean up
    $texData delete
}

proc InitExtensions {} {
    if { ![tcl3dHaveExtension "GL_ARB_vertex_shader"] } {
        error "Extension GL_ARB_vertex_shader missing"
    }
    if { ![tcl3dHaveExtension "GL_ARB_fragment_shader"] } {
        error "Extension GL_ARB_fragment_shader missing"
    }
}

proc ReadShaderFile { fileName } {
    set retVal [catch {open $fileName r} fp]
    if { $retVal == 0 } {
        # Extract the effects parameter values from the first 3 lines of the shader program.
        # The first line is skipped, line 2 and 3 contain the parameter values for the 2 sliders.
        gets $fp dummy
        gets $fp param1
        scan $param1 "// %f %f %f %f" ::g_paramSett(def,1) ::g_paramSett(min,1) \
                                      ::g_paramSett(max,1) ::g_paramSett(inc,1)
        gets $fp param2
        scan $param2 "// %f %f %f %f" ::g_paramSett(def,2) ::g_paramSett(min,2) \
                                      ::g_paramSett(max,2) ::g_paramSett(inc,2)
        # Now read the rest of the file. This should contain the actual shader source.
        set buffer [read $fp]
        close $fp
    } else {
        error "Cannot open shader file $fileName"
    }
    return $buffer
}

proc InitGLSL {} {
    CleanupShader

    set ::g_programDeformObject [glCreateProgramObjectARB]

    set ::g_shaderDeformObject [glCreateShaderObjectARB GL_FRAGMENT_SHADER_ARB]
    glAttachObjectARB $::g_programDeformObject $::g_shaderDeformObject

    set shaderSource [ReadShaderFile "$::g_current(Shader)"]
    set shaderSourceStrings [list $shaderSource]
    set lenList [list [string length $shaderSource]]
    glShaderSourceARB $::g_shaderDeformObject 1 $shaderSourceStrings $lenList

    glCompileShaderARB $::g_shaderDeformObject
    PrintInfoLog $::g_shaderDeformObject

    glLinkProgramARB $::g_programDeformObject
    set success [tcl3dVector GLint 1]
    glGetObjectParameterivARB $::g_programDeformObject GL_OBJECT_LINK_STATUS_ARB $success
    if { [$success get 0] == 0 } {
        error "Shader could not be linked!"
    }
    $success delete

    set ::sampler [glGetUniformLocationARB $::g_programDeformObject "sampler"]
    set ::param1  [glGetUniformLocationARB $::g_programDeformObject "param1"]
    set ::param2  [glGetUniformLocationARB $::g_programDeformObject "param2"]
    set ::texScaleS [glGetUniformLocationARB $::g_programDeformObject "texScaleS"]
    set ::texScaleT [glGetUniformLocationARB $::g_programDeformObject "texScaleT"]

    PrintInfoLog $::g_programDeformObject
}

proc InitTexture {} {
    ReadImg $::g_current(Texture)
    glActiveTexture GL_TEXTURE0
    glBindTexture GL_TEXTURE_2D [$::g_tex get 0]
}

proc tclDisplayFunc { toglwin } {
    glUniform1iARB $::sampler 0                    ; # pass in texture
    glUniform1fARB $::param1 $::g_current(Param,1) ; # pass in parameter1
    glUniform1fARB $::param2 $::g_current(Param,2) ; # pass in parameter2
    glUniform1fARB $::texScaleS $::g_texScaleS ; # pass in parameter2
    glUniform1fARB $::texScaleT $::g_texScaleT ; # pass in parameter2
    
    glClearColor 0.0 0.0 0.0 0.0
    glClear GL_COLOR_BUFFER_BIT
    glBegin GL_QUADS
        glTexCoord2f 0.0 0.0
        glVertex2f 0.0 0.0
        glTexCoord2f $::g_texScaleS 0.0
        glVertex2f $::sizeTexX 0.0
        glTexCoord2f $::g_texScaleS $::g_texScaleT
        glVertex2f $::sizeTexX $::sizeTexY
        glTexCoord2f 0.0 $::g_texScaleT
        glVertex2f 0.0 $::sizeTexY
    glEnd
    $toglwin swapbuffer
}

proc Update { toglwin updateTex } {
    if { $updateTex } {
        InitTexture
    }

    # InitGLSL calls ReadShaderFile, which updates the parameter values in g_paramSett.
    InitGLSL
    glUseProgramObjectARB $::g_programDeformObject

    foreach param { 1 2 } {
        if { $::g_paramSett(inc,$param) == 0 } {
            set ::g_current(Param,$param) 0
            .fr.scale.fr$param.label configure -state disabled
            .fr.scale.fr$param.param configure -state disabled
        } else {
            .fr.scale.fr$param.label configure -state normal
            .fr.scale.fr$param.param configure -state normal
            .fr.scale.fr$param.param configure -from       $::g_paramSett(min,$param)
            .fr.scale.fr$param.param configure -to         $::g_paramSett(max,$param)
            .fr.scale.fr$param.param configure -resolution $::g_paramSett(inc,$param)
            set ::g_current(Param,$param) $::g_paramSett(def,$param)
        }
    }
    $toglwin postredisplay
}

proc tclCreateFunc { toglwin } {
    InitExtensions

    glEnable GL_TEXTURE_2D
    glPolygonMode GL_FRONT GL_FILL
}

proc tclReshapeFunc { toglwin w h } {
    glViewport 0 0 $w $h
    glMatrixMode GL_PROJECTION
    glLoadIdentity
    glOrtho 0.0 $::sizeTexX 0.0 $::sizeTexY 0.0 1.0
    glMatrixMode GL_MODELVIEW
    glLoadIdentity
}

proc CleanupShader {} {
    if { [info exists ::g_shaderDeformObject] } {
        glDeleteObjectARB $::g_shaderDeformObject
    }
    if { [info exists ::g_programDeformObject] } {
        glDeleteObjectARB $::g_programDeformObject
    }
}

proc Cleanup {} {
    if { [info exists ::g_tex] } {
        glDeleteTextures 1 [$::g_tex get 0]
    }
    CleanupShader

    foreach var [info globals g_*] {
        uplevel #0 unset $var
    }
}

proc ExitProg {} {
    exit
}

proc AddTexToList { textureFile } {
    set textureName [file rootname [file tail $textureFile]]
    radiobutton .fr.btns.fr2.b$textureName -text $textureName \
                -variable ::g_current(Texture) -value $textureFile \
                -command "Update .fr.toglwin true"
    pack .fr.btns.fr2.b$textureName -side top -anchor w
}

frame .fr
pack .fr -expand 1 -fill both
togl .fr.toglwin -width 512 -height 512 \
                 -double true -depth true \
                 -createproc  tclCreateFunc \
                 -reshapeproc tclReshapeFunc \
                 -displayproc tclDisplayFunc
frame   .fr.btns
frame   .fr.scale
label   .fr.info
grid .fr.toglwin -row 0 -column 0 -sticky news
grid .fr.btns    -row 0 -column 1 -sticky news -rowspan 2
grid .fr.scale   -row 1 -column 0 -sticky news
grid .fr.info    -row 2 -column 0 -sticky news -columnspan 2
grid rowconfigure .fr 0 -weight 1
grid columnconfigure .fr 0 -weight 1

labelframe .fr.btns.fr1 -text "Effects"
labelframe .fr.btns.fr2 -text "Images"
labelframe .fr.btns.fr3 -text "Actions"
pack .fr.btns.fr1 .fr.btns.fr2 .fr.btns.fr3 -side top -fill x -padx 3 -pady 4
foreach effectFile $g_Effects {
    set effectName [file rootname [file tail $effectFile]]
    radiobutton .fr.btns.fr1.b$effectName -text $effectName \
                -variable g_current(Shader) -value $effectFile \
                -command "Update .fr.toglwin false"
    pack .fr.btns.fr1.b$effectName -side top -anchor w
}
foreach textureFile $g_Textures {
    AddTexToList $textureFile
}
button .fr.btns.fr3.open -text "Load image ..." -command AskOpen
pack .fr.btns.fr3.open -side top -expand 1 -fill x
button .fr.btns.fr3.view -text "View shader" -command ViewShaderFile
pack .fr.btns.fr3.view -side top -expand 1 -fill x

frame .fr.scale.fr1
frame .fr.scale.fr2
pack .fr.scale.fr1 .fr.scale.fr2 -side top -expand 1 -fill x

label .fr.scale.fr1.label -text "Parameter 1:"
label .fr.scale.fr2.label -text "Parameter 2:"
scale .fr.scale.fr1.param -orient horiz -showvalue true -variable g_current(Param,1) \
                          -command { .fr.toglwin postredisplay }
scale .fr.scale.fr2.param -orient horiz -showvalue true -variable g_current(Param,2) \
                          -command { .fr.toglwin postredisplay }
pack .fr.scale.fr1.label .fr.scale.fr2.label -side left
pack .fr.scale.fr1.param .fr.scale.fr2.param -side left -expand 1 -fill x

wm title . "Tcl3D demo: Photo Booth Effects" 

# Watch For ESC Key And Quit Messages
wm protocol . WM_DELETE_WINDOW "ExitProg"
bind . <Key-Escape> "ExitProg"

Update .fr.toglwin true
Update .fr.toglwin true

PrintInfo [format "Running on %s with a %s (OpenGL %s, Tcl %s)" \
           $tcl_platform(os) [glGetString GL_RENDERER] \
           [glGetString GL_VERSION] [info patchlevel]]
