#-----------------------------------------------------------------------------
#           Name: ogl_cg_vertex_displacement.cpp 
#                 ogl_glslang_vertex_displacement.cpp
#         Author: Kevin Harris (kevin@codesampler.com)
#  Last Modified: 04/21/05
#    Description: This sample demonstrates how to perform mesh deformation or 
#                 vertex displacement with OpenGL using a Cg or GLSL shader.
#
#   Control Keys: F1 - Increase flag motion
#                 F2 - Decrease flag motion
#                 F3 - Toggle wire-frame mode
#-----------------------------------------------------------------------------
#
# Original C++ code by Kevin Harris (kevin@codesampler.com)
# See www.codesampler.com for the original files
# OpenGL samples page 11: Vertex Displacement or Mesh Deformation Shader
#
# Modified for Tcl3D by Paul Obermeier 2005/11/05
# See www.tcl3d.org for the Tcl3D extension.
#
# This sample integrates the Cg and GLSL code into one file. 
# If called with no command line arguments, it uses the Cg shader.
# Use "glsl" as parameter to use the GLSL shader.

package require Tk
package require Img
package require tcl3d 0.3.3

# Font to be used in the Tk listbox.
set g_listFont {-family {Courier} -size 10}

set g_WinWidth  640
set g_WinHeight 480

set g_LastMousePosX(1) 0
set g_LastMousePosY(1) 0

set g_fSpinX(1) 0.0
set g_fSpinY(1) 0.0

set g_fCurrentAngle     0.0
set g_fSpeedOfRotation 10.0
set g_bWireFrameMode    0
set g_bUseCgShader      1
set g_bAnimStarted      0

set g_iFrameCount 0

set g_StopWatch [tcl3dNewSwatch]
tcl3dStartSwatch $g_StopWatch
set g_fElapsedTime 0.0
set ::s_lastTime [tcl3dLookupSwatch $::g_StopWatch]

# Mesh properties
set g_nNumVertsAlongX   16
set g_nNumVertsAlongZ   16
set g_fMeshLengthAlongX  8.0
set g_fMeshLengthAlongZ  5.0

# GL_T2F_N3F_V3F
set TU 0
set TV 1
set NX 2
set NY 3
set NZ 4
set X  5
set Y  6
set Z  7
set SIZE 8

set g_nMeshVertCount [expr {($::g_nNumVertsAlongX-1) * \
                            ($::g_nNumVertsAlongZ-1) * 6}]
set g_meshVertices [tcl3dVector GLfloat [expr {$::g_nMeshVertCount * $SIZE}]]

# Determine the directory of this script.
set g_scriptDir [file dirname [info script]]

# Show errors occuring in the Togl callbacks.
proc bgerror { msg } {
    tk_messageBox -icon error -type ok -message "Error: $msg"
    exit
}

# Print info message into widget a the bottom of the window.
proc PrintInfo { msg } {
    if { [winfo exists .fr.info] } {
        .fr.info configure -text $msg
    }
}

proc SetSpeedOfRotation { toglwin off } {
    set ::g_fSpeedOfRotation [expr {$::g_fSpeedOfRotation + $off}]
    $toglwin postredisplay
}

proc ToggleWireframe { toglwin } {
    set ::g_bWireFrameMode [expr {1 - $::g_bWireFrameMode}]
    $toglwin postredisplay
}

proc SetMouseInput { x y } {
    set ::g_LastMousePosX(1) $x
    set ::g_LastMousePosY(1) $y
}

proc GetMouseInput { x y } {
    set nXDiff [expr {($x - $::g_LastMousePosX(1))}]
    set nYDiff [expr {($y - $::g_LastMousePosY(1))}]
        
    set ::g_fSpinX(1) [expr {$::g_fSpinX(1) - $nXDiff}]
    set ::g_fSpinY(1) [expr {$::g_fSpinY(1) - $nYDiff}]

    set ::g_LastMousePosX(1) $x
    set ::g_LastMousePosY(1) $y
    .fr.toglwin postredisplay
}

proc tclReshapeFunc { toglwin w h } {
    set ::g_WinWidth  $w
    set ::g_WinHeight $h
    glViewport 0 0 $w $h
    glMatrixMode GL_PROJECTION
    glLoadIdentity
    gluPerspective 45.0 [expr {double($w)/double($h)}] 0.1 100.0
}

proc tclCreateFunc { toglwin } {
    glEnable GL_TEXTURE_2D
    glEnable GL_DEPTH_TEST
    glClearColor 0.0 0.0 0.0 1.0

    glMatrixMode GL_PROJECTION
    glLoadIdentity
    gluPerspective 45.0 [expr {double($::g_WinWidth)/double($::g_WinHeight)}] \
                   0.1 100.0

    set texName [file join $::g_scriptDir "us_flag.bmp"]
    set retVal [catch {set phImg [image create photo -file $texName]} err1]
    if { $retVal != 0 } {
        error "Error reading image $texName ($err1)"
    } else {
        set w [image width  $phImg]
        set h [image height $phImg]
        set n [tcl3dPhotoChans $phImg]
        set pTextureImage [tcl3dVectorFromPhoto $phImg]
        image delete $phImg
    }

    set ::g_textureID [tcl3dVector GLuint 1]
    glGenTextures 1 $::g_textureID

    glBindTexture GL_TEXTURE_2D [$::g_textureID get 0]

    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER $::GL_LINEAR
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MAG_FILTER $::GL_LINEAR

    if { $n == 3 } {
        set type $::GL_RGB
    } else {
       set type $::GL_RGBA
    }
    glTexImage2D GL_TEXTURE_2D 0 $n $w $h 0 $type GL_UNSIGNED_BYTE $pTextureImage

    $pTextureImage delete

    CreateMesh

    if { $::g_bUseCgShader } {
        InitCgShader
    } else {
        InitGLSLShader
    }
}

proc CreateMesh {} {
    # Compute position deltas for moving down the X, and Z axis
    # during mesh creation
    set dX [expr { (1.0 / ($::g_nNumVertsAlongX-1))}]
    set dZ [expr {-(1.0 / ($::g_nNumVertsAlongZ-1))}]

    # Compute tex-coord deltas for moving down the X, and Z axis
    # during mesh creation
    set dTU [expr {1.0 / ($::g_nNumVertsAlongX-1)}]
    set dTV [expr {1.0 / ($::g_nNumVertsAlongZ-1)}]

    # These are all the same...
    for {set i 0 } { $i < $::g_nMeshVertCount } { incr i } {
        set ind [expr {$i * $::SIZE}]
        # Mesh tesselation occurs in the X,Z plane, so Y is always zero
        $::g_meshVertices set [expr {$ind + $::Y}]  0.0

        $::g_meshVertices set [expr {$ind + $::NX}] 0.0
        $::g_meshVertices set [expr {$ind + $::NY}] 1.0
        $::g_meshVertices set [expr {$ind + $::NZ}] 0.0
    }

    # Create all the vertex points required by the mesh...
    # Note: Mesh tesselation occurs in the X,Z plane.

    # For each row of our mesh...
    set i 0
    for {set z 0 } { $z < [expr {$::g_nNumVertsAlongZ-1}] } { incr z } {
        # Fill the row with quads which are composed of two triangles each...
        for {set x 0 } { $x < [expr {$::g_nNumVertsAlongX-1}] } { incr x } {
            # First triangle of the current quad
            #   ___ 2
            #  |  /|
            #  |/__|
            # 0     1

            # 0
            set off [expr {$i*$::SIZE}]
            $::g_meshVertices set [expr {$off + $::X}]  [expr {$::g_fMeshLengthAlongX * $x * $dX}]
            $::g_meshVertices set [expr {$off + $::Z}]  [expr {$::g_fMeshLengthAlongZ * $z * $dZ}]
            $::g_meshVertices set [expr {$off + $::TU}] [expr {$x * $dTU}]
            $::g_meshVertices set [expr {$off + $::TV}] [expr {$z * $dTV}]
            incr i

            # 1
            set off [expr {$i*$::SIZE}]
            $::g_meshVertices set [expr {$off + $::X}]  [expr {$::g_fMeshLengthAlongX * ($x+1.0) * $dX}]
            $::g_meshVertices set [expr {$off + $::Z}]  [expr {$::g_fMeshLengthAlongZ * $z * $dZ}]
            $::g_meshVertices set [expr {$off + $::TU}] [expr {($x+1.0) * $dTU}]
            $::g_meshVertices set [expr {$off + $::TV}] [expr {$z * $dTV}]
            incr i

            # 2
            set off [expr {$i*$::SIZE}]
            $::g_meshVertices set [expr {$off + $::X}]  [expr {$::g_fMeshLengthAlongX * ($x+1.0) * $dX}]
            $::g_meshVertices set [expr {$off + $::Z}]  [expr {$::g_fMeshLengthAlongZ * ($z+1.0) * $dZ}]
            $::g_meshVertices set [expr {$off + $::TU}] [expr {($x+1.0) * $dTU}]
            $::g_meshVertices set [expr {$off + $::TV}] [expr {($z+1.0) * $dTV}]
            incr i

            # Second triangle of the current quad
            # 2 ___ 1
            #  |  /|
            #  |/__|
            # 0

            # 0
            set off [expr {$i*$::SIZE}]
            $::g_meshVertices set [expr {$off + $::X}]  [expr {$::g_fMeshLengthAlongX * $x * $dX}]
            $::g_meshVertices set [expr {$off + $::Z}]  [expr {$::g_fMeshLengthAlongZ * $z * $dZ}]
            $::g_meshVertices set [expr {$off + $::TU}] [expr {$x * $dTU}]
            $::g_meshVertices set [expr {$off + $::TV}] [expr {$z * $dTV}]
            incr i

            # 1
            set off [expr {$i*$::SIZE}]
            $::g_meshVertices set [expr {$off + $::X}]  [expr {$::g_fMeshLengthAlongX * ($x+1.0) * $dX}]
            $::g_meshVertices set [expr {$off + $::Z}]  [expr {$::g_fMeshLengthAlongZ * ($z+1.0) * $dZ}]
            $::g_meshVertices set [expr {$off + $::TU}] [expr {($x+1.0) * $dTU}]
            $::g_meshVertices set [expr {$off + $::TV}] [expr {($z+1.0) * $dTV}]
            incr i

            # 2
            set off [expr {$i*$::SIZE}]
            $::g_meshVertices set [expr {$off + $::X}]  [expr {$::g_fMeshLengthAlongX * $x * $dX}]
            $::g_meshVertices set [expr {$off + $::Z}]  [expr {$::g_fMeshLengthAlongZ * ($z+1.0) * $dZ}]
            $::g_meshVertices set [expr {$off + $::TU}] [expr {$x * $dTU}]
            $::g_meshVertices set [expr {$off + $::TV}] [expr {($z+1.0) * $dTV}]
            incr i
        }
    }

    set ::g_meshDisplayList [glGenLists 1]
    glNewList $::g_meshDisplayList GL_COMPILE
    glInterleavedArrays GL_T2F_N3F_V3F 0 $::g_meshVertices
    glDrawArrays GL_TRIANGLES 0 $::g_nMeshVertCount
    glEndList
}

proc ReadShaderFile { fileName } {
    set pathName [file join $::g_scriptDir $fileName]
    set realPath [tcl3dGetExtFile $pathName]
    set retVal [catch {open $realPath r} fp]
    if { $retVal == 0 } {
        set buffer [read $fp]
        close $fp
    } else {
        error "Cannot open shader file $realPath"
    }
    return $buffer
}

proc InitGLSLShader {} {
    if { ![tcl3dOglHaveExtension "GL_ARB_shading_language_100"] } {
        error "Extension GL_ARB_shading_language_100 missing"
    }
    if { ![tcl3dOglHaveExtension "GL_ARB_shader_objects"] } {
        error "Extension GL_ARB_shader_objects missing"
    }
    set bLinked       [tcl3dVector GLint 1]
    set bVertCompiled [tcl3dVector GLint 1]

    # Create the vertex shader...
    set ::g_vertexShader [glCreateShaderObjectARB GL_VERTEX_SHADER_ARB]

    set vertexShaderAssembly [ReadShaderFile "ogl_glslang_vertex_displacement.vert"]
    tcl3dOglShaderSource $::g_vertexShader $vertexShaderAssembly
    glCompileShaderARB $::g_vertexShader

    glGetObjectParameterivARB $::g_vertexShader GL_OBJECT_COMPILE_STATUS_ARB \
                              $bVertCompiled
    if { [$bVertCompiled get 0] == 0 } {
        set msg [tcl3dOglGetInfoLogARB $::g_vertexShader]
        error "Vertex Shader Compile Error: $msg"
    }

    # Create a program object and attach our flag shader to it...

    set ::g_programObj [glCreateProgramObjectARB]
    glAttachObjectARB $::g_programObj $::g_vertexShader

    # Link the program object and print out the info log...

    glLinkProgramARB $::g_programObj
    glGetObjectParameterivARB $::g_programObj GL_OBJECT_LINK_STATUS_ARB $bLinked

    if { [$bLinked get 0] == 0 } {
        set msg [tcl3dOglGetInfoLogARB $::g_programObj]
        error "Linking Error: $msg"
    }

    # Locate some parameters by name so we can set them later...

    set ::g_param_currentAngle [glGetUniformLocationARB $::g_programObj "currentAngle"]

    $bLinked delete
    $bVertCompiled delete
}

proc InitCgShader {} {
    # Search for a valid vertex shader profile in this order:
    #
    # CG_PROFILE_ARBVP1 - GL_ARB_vertex_program
    # CG_PROFILE_VP40   - GL_ARB_vertex_program + GL_NV_vertex_program3

    set ::g_CGprofile [tcl3dCgFindProfile $::CG_PROFILE_ARBVP1 \
                                          $::CG_PROFILE_VP40]
    if { $::g_CGprofile eq "" } {
        error "Failed to initialize vertex shader. No ARBVP1 or VP40 profile available."
    }

    # Create a CG context
    set ::g_CGcontext [cgCreateContext]
  
    # Add the Vertex shader to the context
    set cgFile [tcl3dGetExtFile [file join $::g_scriptDir "ogl_cg_vertex_displacement.cg"]]
    set ::g_CGprogram [cgCreateProgramFromFile $::g_CGcontext \
                           CG_SOURCE $cgFile \
                           $::g_CGprofile "main" "NULL"]
    set retVal [tcl3dCgGetError $::g_CGcontext]
    if { $retVal ne "" } {
        error "cgCreateProgramFromFile: $retVal"
    }

    # Load the program using Cg's expanded interface...
    cgGLLoadProgram $::g_CGprogram
    set retVal [tcl3dCgGetError $::g_CGcontext]
    if { $retVal ne "" } {
        error "cgGLLoadProgram: $retVal"
    }

    # Bind some parameters by name so we can set them later...
    set ::g_param_currentAngle [cgGetNamedParameter $::g_CGprogram "currentAngle"]
}

proc SetShaderConstants {} {
    set ::g_fCurrentAngle [expr {$::g_fCurrentAngle - $::g_fSpeedOfRotation * $::g_fElapsedTime}]

    while { $::g_fCurrentAngle > 360.0 } {
        set ::g_fCurrentAngle [expr {$::g_fCurrentAngle - 360.0}]
    }
    while { $::g_fCurrentAngle < 0.0 } {
        set ::g_fCurrentAngle [expr {$::g_fCurrentAngle + 360.0}]
    }

    # Load a value, which continuously grows from 0 to 360 and repeats...
    if { $::g_bUseCgShader } {
        cgGLSetParameter4f $::g_param_currentAngle $::g_fCurrentAngle 0.0 0.0 0.0
    } else {
        if { $::g_param_currentAngle != -1 } {
            glUniform1fARB $::g_param_currentAngle $::g_fCurrentAngle
        }
    }
}

proc tclDisplayFunc { toglwin } {
    set startTime [tcl3dLookupSwatch $::g_StopWatch]

    glClear [expr {$::GL_COLOR_BUFFER_BIT | $::GL_DEPTH_BUFFER_BIT}]

    # Viewport command is not really needed, but has been inserted for
    # Mac OSX. Presentation framework (Tk) does not send a reshape event,
    # when switching from one demo to another.
    glViewport 0 0 $::g_WinWidth $::g_WinHeight

    glMatrixMode GL_MODELVIEW
    glLoadIdentity
    glTranslatef -4.0 -2.5 -10.0
    glRotatef [expr {-1.0 * $::g_fSpinY(1) + 90}] 1.0 0.0 0.0
    glRotatef [expr {-1.0 * $::g_fSpinX(1)}]      0.0 1.0 0.0

    if { $::g_bUseCgShader } {
        SetShaderConstants
        cgGLBindProgram $::g_CGprogram
        cgGLEnableProfile $::g_CGprofile
    } else {
        glUseProgramObjectARB $::g_programObj
        SetShaderConstants
    }

    if { $::g_bWireFrameMode } {
        glPolygonMode GL_FRONT_AND_BACK GL_LINE
    } else {
        glPolygonMode GL_FRONT_AND_BACK GL_FILL
    }

    # Render tesselated mesh...

    if { $::g_bUseCgShader } {
        glBindTexture GL_TEXTURE_2D [$::g_textureID get 0]
    }
    glCallList $::g_meshDisplayList

    if { $::g_bUseCgShader } {
        cgGLDisableProfile $::g_CGprofile
    } else {
        glUseProgramObjectARB 0
    }

    if { $::g_bAnimStarted } {
        DisplayFPS
    }

    $toglwin swapbuffers
    set endTime [tcl3dLookupSwatch $::g_StopWatch]
    set ::g_fElapsedTime [expr {$endTime - $startTime}]
}

proc StartStopAnimation {} {
    if { $::g_bAnimStarted == 0 } {
        StartAnimation
    } else {
        StopAnimation
    }
}

proc StartAnimation {} {
    set ::g_bAnimStarted 1
    .fr.toglwin postredisplay
    set ::animId [after idle StartAnimation]
}

proc StopAnimation {} {
    if { [info exists ::animId] } {
        after cancel $::animId 
        unset ::animId
    }
    set ::g_bAnimStarted 0
    set ::g_iFrameCount 0
}

proc GetFPS { { elapsedFrames 1 } } {
    set currentTime [tcl3dLookupSwatch $::g_StopWatch]
    set fps [expr {$elapsedFrames / ($currentTime - $::s_lastTime)}]
    set ::s_lastTime $currentTime
    return $fps
}

proc DisplayFPS {} {
    incr ::g_iFrameCount
    if { $::g_iFrameCount == 200 } {
        set msg [format "%s (%.0f fps)" $::appTitle [GetFPS $::g_iFrameCount]]
        wm title . $msg
        set ::g_iFrameCount 0
    }
}

proc Cleanup {} {
    if { [info exists ::g_textureID] } {
        glDeleteTextures 1 [$::g_textureID get 0]
        $::g_textureID delete
    }
    if { [info exists ::g_meshDisplayList] } {
        glDeleteLists $::g_meshDisplayList 0
    }
    if { $::g_bUseCgShader } {
        if { [info exists ::g_CGprogram] } {
            cgDestroyProgram $::g_CGprogram
        }
        if { [info exists ::g_CGcontext] } {
            cgDestroyContext $::g_CGcontext
        }
    } else {
        if { [info exists ::g_vertexShader] } {
            glDeleteObjectARB $::g_vertexShader
        }
        if { [info exists ::g_programObj] } {
            glDeleteObjectARB $::g_programObj
        }
    }
    tcl3dDeleteSwatch $::g_StopWatch

    $::g_meshVertices delete

    foreach var [info globals g_*] {
        uplevel #0 unset $var
    }
}

proc ExitProg {} {
    exit
}

if { $argc >= 1 } {
    if { [string compare -nocase [lindex $argv 0] "cg"] == 0 } {
        set ::g_bUseCgShader 1
    } elseif { [string compare -nocase [lindex $argv 0] "glsl"] == 0 } {
        set ::g_bUseCgShader 0
    }
}

if { $::g_bUseCgShader } {
    if { [info procs tcl3dHaveCg] ne "" } {
        if { ![tcl3dHaveCg] } {
            tk_messageBox -icon error -type ok -title "Error" \
                          -message "You do not have Cg installed."
            exit
        }
    }
}

frame .fr
pack .fr -expand 1 -fill both
togl .fr.toglwin -width $g_WinWidth -height $g_WinHeight \
                 -swapinterval 0 \
                 -double true -depth true \
                 -createproc tclCreateFunc \
                 -reshapeproc tclReshapeFunc \
                 -displayproc tclDisplayFunc 
listbox .fr.usage -font $::g_listFont -height 5
label   .fr.info
grid .fr.toglwin -row 0 -column 0 -sticky news
grid .fr.usage   -row 1 -column 0 -sticky news
grid .fr.info    -row 2 -column 0 -sticky news
grid rowconfigure .fr 0 -weight 1
grid columnconfigure .fr 0 -weight 1

set shaderType "GLSL"
if { $::g_bUseCgShader } {
    set shaderType "Cg"
}
set appTitle "Tcl3D demo: CodeSampler's Vertex Displacement Shader using $shaderType"
wm title . $appTitle

# Watch For ESC Key And Quit Messages
wm protocol . WM_DELETE_WINDOW "ExitProg"
bind . <Key-Escape> "ExitProg"
bind . <Key-space>  "StartStopAnimation"
bind . <Key-F1>     "SetSpeedOfRotation .fr.toglwin  0.5"
bind . <Key-F2>     "SetSpeedOfRotation .fr.toglwin -0.5"
bind . <Key-F3>     "ToggleWireframe .fr.toglwin"

bind .fr.toglwin <1>         "SetMouseInput %x %y"
bind .fr.toglwin <B1-Motion> "GetMouseInput %x %y"

.fr.usage insert end "Key-Escape Exit"
.fr.usage insert end "Key-Space  Start|Stop Animation"
.fr.usage insert end "Key-F1|F2  Increase|Decrease speed"
.fr.usage insert end "Key-F3     Toggle wireframe"
.fr.usage insert end "Command line parameters: glsl or cg"
.fr.usage configure -state disabled

PrintInfo [format "Running on %s with a %s (OpenGL %s, Tcl %s)" \
           $tcl_platform(os) [glGetString GL_RENDERER] \
           [glGetString GL_VERSION] [info patchlevel]]
